/**
 * Copyright (C) 2010 Cloudfarming <info@cloudfarming.nl>
 *
 * Licensed under the Eclipse Public License - v 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.eclipse.org/legal/epl-v10.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.cloudfarming.client.sensor.area;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.io.ParseException;
import com.vividsolutions.jts.io.WKTReader;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import nl.cloudfarming.client.logging.AppLogFactory;
import nl.cloudfarming.client.logging.AppLogger;
import nl.cloudfarming.client.model.Plot;
import nl.cloudfarming.client.model.SensorData;
import nl.cloudfarming.client.model.SensorDataFile;
import nl.cloudfarming.client.model.SensorRun;
import nl.cloudfarming.client.sensor.SensorEventKey;
import nl.cloudfarming.eventbus.BaseGuiEventProducer;
import nl.cloudfarming.eventbus.GuiEventProducer;
import org.netbeans.api.progress.ProgressHandle;
import org.netbeans.api.progress.ProgressHandleFactory;
import org.openide.util.Exceptions;
import org.openide.util.NbBundle;

/**
 *
 * @author Timon Veenstra
 */
public class LinkSensorToPlotTask implements Runnable {

    private static final String LOG_MESSAGE_START = "link.message.info.start";
    private static final String LOG_MESSAGE_END = "link.message.info.end";
    private static final String LOG_MESSAGE_NO_PLOTS = "link.message.info.no.plots";
    private static final AppLogger LOG = AppLogFactory.getLogger(LinkSensorToPlotTask.class);
    private static final String PROCESS_TITLE = NbBundle.getMessage(LinkSensorToPlotTask.class, "link.process.title");
    private SensorAreaService service = new SensorAreaService();
    private final ExecutorService pool = Executors.newFixedThreadPool(POOL_SIZE);
    private static ProgressHandle globalProgressHandle;
    private static int globalProgress = 0;
    private final List<Plot> plots = service.findAllPlots();
    private static final int CACHE_SIZE = 1000;
    private static final int POOL_SIZE = 2;
    private final GuiEventProducer eventProducer = new EventProducer();


    @Override
    public void run() {
        LOG.info(LOG_MESSAGE_START);
        if(!plots.isEmpty()){
            globalProgressHandle = ProgressHandleFactory.createHandle(PROCESS_TITLE);
            globalProgressHandle.start(plots.size());

            for (Plot plot : plots) {
                pool.submit(new LinkPlotJob(plot));
            }
        } else {
            LOG.info(LOG_MESSAGE_NO_PLOTS);
        }

    }

    private SensorRun getSensorRun(Plot plot, SensorDataFile<?> sensorDataFile) {
        SensorRun run = service.findSensorRunForPlotAndFile(plot, sensorDataFile);
        if (run == null) {
            run = new SensorRun(sensorDataFile, plot);
            service.persist(run);
        }
        return run;
    }

    class LinkPlotJob implements Runnable {

        private final Plot plot;

        public LinkPlotJob(Plot plot) {
            this.plot = plot;
        }

        @Override
        public void run() {
            int progress = 0;
            ProgressHandle progressHandle = ProgressHandleFactory.createHandle(NbBundle.getMessage(LinkSensorToPlotTask.class, "link.process.plot.title", plot.getName()));
            List<SensorData> sensorDataCache = new ArrayList<SensorData>();
            try {
                GeometryFactory gf = new GeometryFactory();
                WKTReader wKTReader = new WKTReader(gf);

                Geometry plotGeom = wKTReader.read(plot.getWKT());

                Coordinate firstPoint = plotGeom.getEnvelope().getCoordinate();

                double minx = firstPoint.x;
                double maxx = firstPoint.x;
                double miny = firstPoint.y;
                double maxy = firstPoint.y;

                for (Coordinate c : plotGeom.getEnvelope().getCoordinates()) {
                    minx = (minx > c.x) ? c.x : minx;
                    maxx = (maxx < c.x) ? c.x : maxx;
                    miny = (miny > c.y) ? c.y : miny;
                    maxy = (maxy < c.y) ? c.y : maxy;
                }

                List<SensorData> sensorData = service.findUnlinkedSensorDataInBox(minx, maxx, miny, maxy);
                progressHandle.start(sensorData.size());
                for (SensorData sd : sensorData) {
                    Geometry point = gf.createPoint(new Coordinate(sd.getLongitude(), sd.getLatitude()));
                    //
                    // check intersection with real plot geometry
                    //
                    if (plotGeom.intersects(point)) {
                        sd.setSensorRun(getSensorRun(plot, sd.getDataSensorFile()));
                        service.persist(sd);
                        sensorDataCache.add(sd);
                        if (sensorDataCache.size() % CACHE_SIZE == 0) {
                            service.persist(sensorDataCache);
                            sensorDataCache.clear();
                        }
                    }
                    progressHandle.progress(++progress);
                }
                if (!sensorDataCache.isEmpty()) {
                    service.persist(sensorDataCache);
                    sensorDataCache.clear();
                }
                progressHandle.finish();
                sensorData.clear();

                globalProgressHandle.progress(++globalProgress);
                if (globalProgress >= plots.size()) {
                    globalProgressHandle.finish();
                    eventProducer.triggerEvent(SensorEventKey.SENSOR_DATA_LINKED);
                    LOG.info(LOG_MESSAGE_END);
                }
            } catch (ParseException ex) {
                Exceptions.printStackTrace(ex);
            }
        }
    }

    private class EventProducer extends BaseGuiEventProducer {

        public EventProducer() {
            super(SensorEventKey.class);
        }
    }
}
